@@ -144,7 +144,7 @@
 static const char copyright[] =
     "Copyright (c) 1996-2000\nSleepycat Software Inc.  All rights reserved.\n";
 static const char revid[] =
-    "$Id: htdb_stat.cc,v 1.12 2000/07/21 14:21:24 loic Exp $";
+    "$Id: htdb_stat.cc,v 1.13 2001/02/07 17:21:33 loic Exp $";
 #endif
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -168,6 +168,10 @@ static const char revid[] =
 #include <unistd.h>
 #endif
 
+#ifdef HAVE_GETOPT_H
+#include <getopt.h>
+#endif /* HAVE_GETOPT_H */
+
 extern "C" {
 #include "db_int.h"
 #include "db_page.h"
