@@ -1,11 +1,11 @@
 //
 // Part of the ht://Dig package   <http://www.htdig.org/>
-// Copyright (c) 1999, 2000 The ht://Dig Group
+// Copyright (c) 1999, 2000, 2001 The ht://Dig Group
 // For copyright details, see the file COPYING in your distribution
 // or the GNU General Public License version 2 or later
 // <http://www.gnu.org/copyleft/gpl.html>
 //
-// $Id: WordTree.cc,v 1.4 2000/11/09 08:17:21 loic Exp $
+// $Id: WordTree.cc,v 1.10 2001/06/29 14:14:08 loic Exp $
 //
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -31,6 +31,61 @@ static char* operator_name[WORD_TREE_OP_SIZE] = {
   0
 };
 
+// ************************* WordTree implementation ****************
+
+int WordTree::WalkNextExclude(const WordKey& key)
+{
+  int exclude = 0;
+  if(results) {
+    exclude = results->Exists(key);
+  }
+  return exclude;
+}
+
+int WordTree::TopLevelOptimize(WordTree*& expr)
+{
+  int verbose = expr->verbose;
+  if(verbose) fprintf(stderr, "WordSearch::Optimize: non optimized expression %s\n", (char*)expr->Get());
+  if(expr->Optimize() != OK)
+    return NOTOK;
+
+  /*
+   * Single mandatory word equivalent to single word. 
+   */
+  if(expr->IsA() == WORD_TREE_MANDATORY) {
+    WordTreeMandatory* mandatory = (WordTreeMandatory*)expr;
+    if(mandatory->cursors_length != 1) {
+      fprintf(stderr, "WordTree::TopLevelOptimize: unexpected cursor length == %d instead of 1\n", mandatory->cursors_length);
+      return NOTOK;
+    }
+    expr = mandatory->cursors[0];
+    mandatory->cursors_length = 0;
+    delete mandatory;
+  }
+
+  /* 
+   * Single forbiden word is something we don't want to handle.
+   */
+  if(expr->IsA() == WORD_TREE_NOT) {
+    delete expr;
+    expr = 0;
+  }
+  
+  if(verbose) {
+    if(expr) {
+      fprintf(stderr, "WordTree::TopLevelOptimize: optimized expression %s\n", (char*)expr->Get());
+    } else {
+      fprintf(stderr, "WordTree::TopLevelOptimize: search expression is meaningless, discarded\n");
+    }
+  }
+
+  if(expr && expr->IsA() != WORD_TREE_LITERAL && expr->key_semantic.HasRealm()) {
+    if(expr->ClearRealm() != OK)
+      return NOTOK;
+  }
+  return OK;
+}
+
 // ************************* WordTreeLiteral implementation ****************
 
 int WordTreeLiteral::WalkInit()
@@ -45,7 +100,6 @@ int WordTreeLiteral::WalkInit()
 
 int WordTreeLiteral::WalkRewind()
 {
-  current_document.Clear();
   if(WordCursorOne::WalkRewind() != OK) return NOTOK;
   if(!bottom_document.Empty())
     return Seek(bottom_document);
@@ -54,29 +108,19 @@ int WordTreeLiteral::WalkRewind()
 
 int WordTreeLiteral::WalkNext()
 {
-  int ret;
-  do {
-    ret = WordCursorOne::WalkNext();
-    if(verbose > 3) fprintf(stderr, "WordTreeLiteral::WalkNext: reached %s\n", (char*)GetDocument().Get());
-  } while(ret == OK &&
-	  key_semantic.DocumentCompare(current_document, GetDocument()) == 0);
+  int ret = WordCursorOne::WalkNext();
+  if(verbose > 3) fprintf(stderr, "WordTreeLiteral::WalkNext: reached %s\n", (char*)GetDocument().Get());
 
   if(ret == OK && !top_document.Empty()) {
     if(key_semantic.DocumentCompare(top_document, GetDocument()) <= 0)
       ret = WORD_WALK_ATEND;
   }
-  
-  if(ret == OK) {
-    current_document = GetDocument();
-  } else
-    current_document.Clear();
 
   return ret;
 }
 
 int WordTreeLiteral::Seek(const WordKey& position)
 {
-  current_document.Clear();
   return WordCursorOne::Seek((bottom_document.Empty() || key_semantic.DocumentCompare(bottom_document, position) <= 0) ? position : bottom_document);
 }
 
@@ -113,6 +157,9 @@ static char* ret2str(int ret)
   if(ret == WORD_WALK_ATEND)
     return "ATEND";
 
+  if(ret == WORD_WALK_ATEND|WORD_WALK_ATEND_NOMATCH)
+    return "ATEND|NOMATCH";
+
   return "???";
 }
 
@@ -228,7 +275,7 @@ WordTreeOperand::Seek(const WordKey& patch)
   int ret = OK;
   for(i = 0; i < cursors_length; i++)
     if((ret = cursors[i]->Seek(patch)) != OK &&
-       ret != WORD_WALK_ATEND)
+       !cursors[i]->IsAtEnd())
       return ret;
   status = OK;
   return OK;
@@ -258,10 +305,11 @@ int WordTreeOptional::Optimize()
   }
 
   unsigned int stripped;
-  if((ret = StripNonExistent(stripped)) != OK)
+  unsigned int killed_mandatory;
+  if((ret = StripNonExistent(stripped, killed_mandatory)) != OK)
     return ret;
 
-  if(AllOrNothing() && stripped) {
+  if(killed_mandatory || (AllOrNothing() && stripped)) {
     //
     // One word is missing and everything is lost,
     // Just kill the remaining cursors.
@@ -338,142 +386,136 @@ int WordTreeOptional::WalkNext()
     if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: required position or document is reset, i.e. before first possible entry\n");
     key_semantic.DocumentClear(constraint);
   }
-  
-  //
-  // Advance cursors to next constraint, if not at the
-  // beginning of the search.
-  //
+
   int ret = OK;
-  int match_ok = 0;
-  do {
-    //
-    // Advance cursors so that next call fetches another constraint
-    //
-    if(cursor_get_flags == DB_NEXT)
-      key_semantic.DocumentNext(constraint, uniq);
+  //
+  // Advance cursors so that next call fetches another constraint
+  //
+  if(cursor_get_flags == DB_NEXT)
+    key_semantic.DocumentNext(constraint, uniq);
     
-    if((ret = Seek(constraint)) != OK)
-      return ret;
+  if((ret = Seek(constraint)) != OK)
+    return ret;
 
-    int near = permutation.Proximity();
-    WordTree* first = 0;
-    for(unsigned int i = 0; i < cursors_length;) {
-      WordTree& cursor = *(cursors[i]);
-      near = permutation.Proximity();
-      int excluded = permutation.Excluded(i);
-      if(verbose) fprintf(stderr, "\n  WordTreeOptional::WalkNext: looking for %s (excluded = %s, proximity = %s)\n", (char*)cursor.search, (excluded ? "yes" : "no"), (near ? "yes" : "no" ));
+  int near = permutation.Proximity();
+  WordTree* first = 0;
+  for(unsigned int i = 0; i < cursors_length;) {
+    int excluded = permutation.Excluded(i);
+    WordTree& cursor = *(cursors[i]);
 
-      int ret;
-      if(excluded) {
-	ret = SearchCursorNot(cursor, constraint);
-	if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: Not -> %s\n", ret2str(ret));
+    if(excluded && !permutation.Ignored(i)) {
+      if(verbose) fprintf(stderr, "\n  WordTreeOptional::WalkNext: looking for %s (ignored)\n", (char*)cursor.search);
+      i++;
+      continue;
+    }
+
+    near = permutation.Proximity();
+    if(verbose) fprintf(stderr, "\n  WordTreeOptional::WalkNext: looking for %s (excluded = %s, proximity = %s)\n", (char*)cursor.search, (excluded ? "yes" : "no"), (near ? "yes" : "no" ));
+
+    int ret;
+    if(excluded) {
+      ret = SearchCursorNot(cursor, constraint);
+      if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: Not -> %s\n", ret2str(ret));
+    } else {
+      if(near) {
+	ret = SearchCursorNear(cursor, first, constraint, proximity);
+	if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: Near -> %s\n", ret2str(ret));
       } else {
-	if(near) {
-	  ret = SearchCursorNear(cursor, first, constraint, proximity);
-	  if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: Near -> %s\n", ret2str(ret));
-	} else {
-	  ret = SearchCursorAnd(cursor, constraint, permutation);
-	  if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: And -> %s\n", ret2str(ret));
-	}
+	ret = SearchCursorAnd(cursor, constraint, permutation);
+	if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: And -> %s\n", ret2str(ret));
       }
+    }
 
-      switch(ret) {
-      case WORD_WALK_ATEND:
-	if(UsePermutation()) {
+    switch(ret & WORD_WALK_RESULT_MASK) {
+    case WORD_WALK_ATEND:
+      //
+      // If a realm exists between the word and the document and the
+      // cursor did not hit the end of the index.
+      //
+      if(key_semantic.HasRealm()) {
+	if(ret & WORD_WALK_ATEND_NOMATCH) {
 	  //
-	  // The search is over with this permutation, try another one.
+	  // If the cursor is still on the same word
 	  //
-	  if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: try next proximity/exclusion permutation\n");
-	  switch(permutation.Next()) {
+	  if(cursor.GetFound().Key().Get(WORD_KEY_WORD) == cursor.GetSearch().Get(WORD_KEY_WORD)) {
 	    //
-	    // No permutations left, the end
+	    // Change the realm of all cursors
 	    //
-	  case WORD_PERMUTE_END:
-	    if(verbose) fprintf(stderr, "\nWordTreeOptional::WalkNext: ATEND\n");
-	    return (status = WORD_WALK_ATEND);
-	    break;
-
+	    if(SetRealm(cursor.GetFound().Key()) != OK)
+	      return NOTOK;
 	    //
-	    // Sart over with this permutation
+	    // And restart the search in this realm
 	    //
-	  case WORD_PERMUTE_OK:
 	    if(WalkRewind() != OK)
 	      return NOTOK;
+	    first = 0;
+	    i = 0;
 	    break;
 	  }
-	  first = 0;
-	  i = 0;
-	} else {
+	}
+      }
+      if(UsePermutation()) {
+	//
+	// The search is over with this permutation, try another one.
+	//
+	if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: try next proximity/exclusion permutation\n");
+	switch(permutation.Next()) {
+	  //
+	  // No permutations left, the end
+	  //
+	case WORD_PERMUTE_END:
 	  if(verbose) fprintf(stderr, "\nWordTreeOptional::WalkNext: ATEND\n");
 	  return (status = WORD_WALK_ATEND);
+	  break;
+
+	  //
+	  // Sart over with this permutation
+	  //
+	case WORD_PERMUTE_OK:
+	  if(key_semantic.HasRealm() && ClearRealm() != OK)
+	    return NOTOK;
+	  if(WalkRewind() != OK)
+	    return NOTOK;
+	  break;
 	}
-	break;
-      case WORD_WALK_REDO:
-	break;
-      case WORD_WALK_RESTART:
 	first = 0;
 	i = 0;
-	break;
-      case WORD_WALK_NEXT:
-	i++;
-	break;
-      case NOTOK:
-      default:
-	if(verbose) fprintf(stderr, "\nWordTreeOptional::WalkNext: NOTOK\n");
-	return ret;
-	break;
+      } else {
+	if(verbose) fprintf(stderr, "\nWordTreeOptional::WalkNext: ATEND\n");
+	return (status = WORD_WALK_ATEND);
       }
+      break;
+    case WORD_WALK_REDO:
+      break;
+    case WORD_WALK_RESTART:
+      first = 0;
+      i = 0;
+      break;
+    case WORD_WALK_NEXT:
+      i++;
+      break;
+    case NOTOK:
+    default:
+      if(verbose) fprintf(stderr, "\nWordTreeOptional::WalkNext: %s\n", ret2str(ret));
+      return ret;
+      break;
     }
+  }
 
-    cursor_get_flags = DB_NEXT;
+  cursor_get_flags = DB_NEXT;
 
-    SetInfo();
+  SetInfo();
     
-    //
-    // Save possible result, i.e. first non excluded cursor
-    //
-    for(unsigned int i = 0; i < cursors_length; i++) {
-      WordTree& cursor = *(cursors[i]);
-      if(!permutation.Excluded(i)) {
-	found.Key() = cursor.GetFound().Key();
-	break;
-      }
+  //
+  // Save possible result, i.e. first non excluded cursor
+  //
+  for(unsigned int i = 0; i < cursors_length; i++) {
+    WordTree& cursor = *(cursors[i]);
+    if(!permutation.Excluded(i)) {
+      found.Key() = cursor.GetFound().Key();
+      break;
     }
-
-    match_ok = 1;
-    //
-    // Only bother if near and non near search are involved
-    //
-    if(UseProximity() == WORD_PERMUTE_PROXIMITY_TOGGLE) {
-      //
-      // If we reach this point in the function and 
-      // either proximity search is active or there is
-      // only one word involved, the match is valid. 
-      // Otherwise it may be excluded, see below.
-      //
-      if(!near && permutation.NotExcludedCount() > 1) {
-	//
-	// If not using proximity, a match that fits the proximity 
-	// requirements must be skipped because it was matched by 
-	// the previous permutation (see WordPermute).
-	//
-	switch(CursorsObeyProximity(constraint)) {
-	case OK:
-	  if(verbose) fprintf(stderr, "  WordTreeOptional::WalkNext: match discarded because already returned with proximity constraints\n");
-	  match_ok = 0;
-	  break;
-	case WORD_SEARCH_NOPROXIMITY:
-	  match_ok = 1;
-	  break;
-	default:
-	case NOTOK:
-	  if(verbose) fprintf(stderr, "\nWordTreeOptional::WalkNext: NOTOK\n");
-	  return NOTOK;
-	  break;
-	}
-      }
-    }
-  } while(!match_ok && ret == OK);
+  }
 
   if(verbose) fprintf(stderr, "\nWordTreeOptional::WalkNext: OK\n");
   return ret;
@@ -568,7 +610,7 @@ int WordTreeOptional::SearchCursorNear(WordTree& cursor, WordTree*& master, Word
     return WORD_WALK_NEXT;
 
     //
-    // If current location is above cursor location
+    // If current location is after cursor location
     //
   } else if(direction > 0) {
     //
@@ -631,64 +673,97 @@ int WordTreeOptional::SearchCursorNear(WordTree& cursor, WordTree*& master, Word
 
 int WordTreeOptional::SearchCursorNot(WordTree& cursor, WordKey& document)
 {
-  int direction = key_semantic.DocumentCompare(document, cursor.GetFound().Key());
-  if(verbose) {
-    fprintf(stderr, "  WordTreeOptional::SearchCursorNot: required document is %s\n", (char*)(document.Get()));
-    fprintf(stderr, "  WordTreeOptional::SearchCursorNot: cursor %s is %s\n",
-	    (cursor.IsAtEnd() ? "" : (char*)(cursor.GetFound().Key().Get())),
-	    (cursor.IsAtEnd() ? "at end" : ((direction == 0 ? "ok" : (direction < 0 ? "after required document" : "before required document")))));
-  }
-
-  //
-  // If the cursor is above the current document
-  // (being at the end of walk is being above all documents).
-  //
-  // Means that the cursor is positioned in an acceptable document
-  // and proceed to the next cursor.
-  //
-  if(direction < 0 || cursor.IsAtEnd()) {
-    if(verbose) fprintf(stderr, "  WordTreeOptional::SearchCursorNot: cursor does not match in required document\n");
-    return WORD_WALK_NEXT;
-
+  if(key_semantic.HasRealm()) {
     //
-    // If the cursor is below current document
-    //
-  } else if(direction > 0) {
-    //
-    // Move the cursor up to the document
+    // Since there is a realm we cannot rely on the sequential ordering
+    // of the entries for a given document and we have to start over
+    // each time.
     //
+    WordKey search = cursor.GetSearch();
+    key_semantic.DocumentCopy(document, cursor.GetSearch());
+    cursor.UndefinedRealm();
+    cursor.WalkRewind();
     cursor.Seek(document);
     if(verbose) fprintf(stderr, "  WordTreeOptional::SearchCursorNot: seek to required document\n");
-    int ret;
-    if((ret = cursor.WalkNext()) != OK && ret != WORD_WALK_ATEND)
+    if(cursor.WalkNext() != OK && !cursor.IsAtEnd())
       return NOTOK;
-    //
-    // It is expected in this case that the cursor has moved above
-    // the current document and another visit in the loop will
-    // tell us.
-    //
-    return WORD_WALK_REDO;
+    int ret;
+    if(cursor.IsAtEnd()) {
+      ret = WORD_WALK_NEXT;
+    } else {
+      if(verbose) fprintf(stderr, "  WordTreeOptional::SearchCursorNot: required document incremented\n");
+      //
+      // The cursor does not give any hint on a possible
+      // next document, just go to the next possible one.
+      //
+      key_semantic.DocumentNext(document, uniq);
+      //
+      // Since the current document changed, start over.
+      //
+      ret = WORD_WALK_RESTART;
+    }
+    cursor.GetSearch() = search;
+    cursor.WalkRewind();
+    return ret;
+  } else {
+    int direction = key_semantic.DocumentCompare(document, cursor.GetFound().Key());
+    if(verbose) {
+      fprintf(stderr, "  WordTreeOptional::SearchCursorNot: required document is %s\n", (char*)(document.Get()));
+      fprintf(stderr, "  WordTreeOptional::SearchCursorNot: cursor %s is %s\n",
+	      (cursor.IsAtEnd() ? "" : (char*)(cursor.GetFound().Key().Get())),
+	      (cursor.IsAtEnd() ? "at end" : ((direction == 0 ? "ok" : (direction < 0 ? "after required document" : "before required document")))));
+    }
 
     //
-    // If the cursor matches the current document.
+    // If the cursor is after the current document
+    // (being at the end of walk is being after all documents).
     //
-    // Means that the current document is not a possible match
-    // since it is pointed by this cursor.
+    // Means that the cursor is positioned in an acceptable document
+    // and proceed to the next cursor.
     //
-  } else if(direction == 0) {
-    if(verbose) fprintf(stderr, "  WordTreeOptional::SearchCursorNot: required document incremented\n");
-    //
-    // The cursor does not give any hint on a possible
-    // next document, just go to the next possible one.
-    //
-    key_semantic.DocumentNext(document, uniq);
-    //
-    // Since the current document changed, start over.
-    //
-    return WORD_WALK_RESTART;
-  } else {
-    fprintf(stderr, "  WordTreeOptional::WordCursorNot: reached unreachable statement\n");
-    return NOTOK;
+    if(direction < 0 || cursor.IsAtEnd()) {
+      if(verbose) fprintf(stderr, "  WordTreeOptional::SearchCursorNot: cursor does not match in required document\n");
+      return WORD_WALK_NEXT;
+
+      //
+      // If the cursor is before current document
+      //
+    } else if(direction > 0) {
+      //
+      // Move the cursor up to the document
+      //
+      cursor.Seek(document);
+      if(verbose) fprintf(stderr, "  WordTreeOptional::SearchCursorNot: seek to required document\n");
+      if(cursor.WalkNext() != OK && !cursor.IsAtEnd())
+	return NOTOK;
+      //
+      // It is expected in this case that the cursor has moved after
+      // the current document and another visit in the loop will
+      // tell us.
+      //
+      return WORD_WALK_REDO;
+
+      //
+      // If the cursor matches the current document.
+      //
+      // Means that the current document is not a possible match
+      // since it is pointed by this cursor.
+      //
+    } else if(direction == 0) {
+      if(verbose) fprintf(stderr, "  WordTreeOptional::SearchCursorNot: required document incremented\n");
+      //
+      // The cursor does not give any hint on a possible
+      // next document, just go to the next possible one.
+      //
+      key_semantic.DocumentNext(document, uniq);
+      //
+      // Since the current document changed, start over.
+      //
+      return WORD_WALK_RESTART;
+    } else {
+      fprintf(stderr, "  WordTreeOptional::WordCursorNot: reached unreachable statement\n");
+      return NOTOK;
+    }
   }
   return NOTOK;
 }
@@ -713,14 +788,14 @@ int WordTreeOptional::SearchCursorAnd(WordTree& cursor, WordKey& document, WordE
     return WORD_WALK_NEXT;
 
     //
-    // If the cursor is below current document
+    // If the cursor is before current document
     //
   } else if(direction > 0) {
     //
     // Move the cursor up to the document
     //
-    cursor.Seek(document);
     if(verbose) fprintf(stderr, "  WordTreeOptional::SearchCursorAnd: seek to required document\n");
+    cursor.Seek(document);
     int ret;
     if((ret = cursor.WalkNext()) == OK)
       return WORD_WALK_REDO;
@@ -728,11 +803,11 @@ int WordTreeOptional::SearchCursorAnd(WordTree& cursor, WordKey& document, WordE
       return ret;
 
     //
-    // If the cursor is above current document.
+    // If the cursor is after current document.
     //
     // Means the the current document is not a possible match
     // since it will never reach it because it's already
-    // above it.
+    // after it.
     //
   } else if(direction < 0) {
     //
@@ -752,116 +827,6 @@ int WordTreeOptional::SearchCursorAnd(WordTree& cursor, WordKey& document, WordE
   return NOTOK;
 }
 
-int WordTreeOptional::CursorsObeyProximity(WordKey& document)
-{
-  //
-  // Run if more than one word is involved, proximity
-  // is always true if there is only one word.
-  //
-  if(permutation.NotExcludedCount() <= 1) return OK;
-
-  WordKey location(words->GetContext());
-
-  //
-  // The first non excluded cursor contains anchor location.
-  //
-  unsigned int master_index = 0;
-  for(unsigned int i = 0; i < cursors_length; i++) {
-    if(!permutation.Excluded(i)) {
-      master_index = i;
-      break;
-    }
-  }
-  const WordKey& masterKey = cursors[master_index]->GetFound().Key();
-  key_semantic.DocumentSet(masterKey, location);
-
-  for(unsigned int i = 0; i < cursors_length;) {
-    if(permutation.Excluded(i)) {
-      i++;
-      continue;
-    }
-    
-    WordTree& cursor = *(cursors[i]);
-    if(cursor.IsAtEnd()) return WORD_SEARCH_NOPROXIMITY;
-    //    if(cursor.status & WORD_WALK_FAILED) return NOTOK;
-
-    //
-    // If the cursor moved outside of the tested document,
-    // no proximity match is possible.
-    //
-    if(key_semantic.DocumentCompare(cursor.GetFound().Key(), document) != 0)
-      return WORD_SEARCH_NOPROXIMITY;
-
-    int direction = key_semantic.LocationCompare(location, cursor.GetFound().Key(), proximity);
-    
-    //
-    // If the cursor is in the authorized locations, consider
-    // next cursor
-    //
-    if(direction == 0) {
-      //
-      // master cursor makes the rules for location : its location
-      // is the base to calculate other words mandatory loacations.
-      //
-      if(i == master_index)
-	key_semantic.LocationSet(cursor.GetFound().Key(), location);
-      //
-      // Fix location constraint to accomodate proximity tolerance.
-      //
-      key_semantic.LocationNearLowest(location, proximity);
-      i++;
-
-    //
-    // If current location is greater than cursor location
-    //
-    } else if(direction > 0) {
-      //
-      // Move the cursor up to the location.
-      //
-      cursor.Seek(location);
-      if(verbose > 1) fprintf(stderr, "  WordTreeOptional::CursorsObeyProximity: seek to %s\n", (char*)location.Get());
-      int ret;
-      if((ret = cursor.WalkNext()) != OK) {
-	if(ret == WORD_WALK_ATEND) {
-	  return WORD_SEARCH_NOPROXIMITY;
-	} else {
-	  return NOTOK;
-	}
-      }
-      //
-      // Remove the location constraint for the master word
-      // so that it matches and then enforce location for other
-      // keys.
-      //
-      if(i == master_index)
-	key_semantic.Location2Document(location);
-      //
-      // Reconsider the situation for this cursor
-      //
-
-    //
-    // If current location is lower than cursor location,
-    // meaning that the cursor found no match in the current
-    // document.
-    //
-    } else if(direction < 0) {
-      //
-      // Move to next master key, if possible.
-      //
-      if(i != master_index) {
-	key_semantic.LocationSet(masterKey, location);
-	key_semantic.LocationNext(location);
-      }
-      //
-      // Since the current location changed, start over.
-      //
-      i = 0;
-    }
-  }
-
-  return OK;
-}
-
 //
 // Helper class for AscendingFrequency method
 //
@@ -914,9 +879,10 @@ int WordTreeOptional::AscendingFrequency()
   return OK;
 }
 
-int WordTreeOptional::StripNonExistent(unsigned int& stripped)
+int WordTreeOptional::StripNonExistent(unsigned int& stripped, unsigned int& killed_mandatory)
 {
   stripped = 0;
+  killed_mandatory = 0;
 
   WordTree** tmp = new WordTree*[cursors_length];
   memset((char*)tmp, '\0', sizeof(WordTree*[cursors_length]));
@@ -935,6 +901,8 @@ int WordTreeOptional::StripNonExistent(unsigned int& stripped)
     if(frequency > 0) {
       tmp[to++] = cursors[from];
     } else {
+      if(cursors[from]->IsA() == WORD_TREE_MANDATORY)
+	killed_mandatory++;
       delete cursors[from];
       stripped++;
     }
@@ -964,7 +932,8 @@ int WordTreeOr::Optimize()
     return ret;
 
   unsigned int stripped;
-  if((ret = StripNonExistent(stripped)) != OK)
+  unsigned int killed_mandatory;
+  if((ret = StripNonExistent(stripped, killed_mandatory)) != OK)
     return ret;
 
   return OptimizeOr(WORD_PERMUTE_PROXIMITY_NO);
@@ -1075,15 +1044,15 @@ int WordTreeOr::WalkNext()
 	      continue;
 	  WordTree& cursor = *(cursors[i]);
 
-	  switch((ret = cursor.WalkNext())) {
+	  switch((ret = cursor.WalkNext()) & WORD_WALK_RESULT_MASK) {
 	  case WORD_WALK_ATEND:
 	      //
-	      // Constraint is above all matches for this cursor
+	      // Constraint is after all matches for this cursor
 	      //
 	      break;
 	  case OK:
 	      //
-	      // If candidate is not set or current cursor is below
+	      // If candidate is not set or current cursor is before
 	      // the current candidate, the curent cursor document becomes
 	      // the candidate.
 	      //
@@ -1120,7 +1089,7 @@ int WordTreeOr::WalkNext()
 	    continue;
 	  WordTree& cursor = *(cursors[i]);
 
-	  switch((ret = cursor.WalkNext())) {
+	  switch((ret = cursor.WalkNext()) & WORD_WALK_RESULT_MASK) {
 	  case WORD_WALK_ATEND:
 	    //
 	    // This excluded cursor can't match the candidate, fine.
