@@ -1,11 +1,11 @@
 //
 // Part of the ht://Dig package   <http://www.htdig.org/>
-// Copyright (c) 1999, 2000 The ht://Dig Group
+// Copyright (c) 1999, 2000, 2001 The ht://Dig Group
 // For copyright details, see the file COPYING in your distribution
 // or the GNU General Public License version 2 or later
 // <http://www.gnu.org/copyleft/gpl.html>
 //
-// $Id: WordTree.h,v 1.3 2000/11/09 07:49:26 loic Exp $
+// $Id: WordTree.h,v 1.9 2001/06/29 14:14:08 loic Exp $
 //
 //
 // NAME
@@ -40,10 +40,12 @@
 #include <WordKeySemantic.h>
 #include <WordPermute.h>
 #include <WordCursorOne.h>
+#include <WordResults.h>
 
 #define WORD_WALK_REDO		0x1000
 #define WORD_WALK_RESTART	0x2000
 #define WORD_WALK_NEXT		0x4000
+#define WORD_WALK_END_CACHE	0x8000
 
 //
 // Return values of CursorsObeyProximity method
@@ -70,6 +72,27 @@
 #define WORD_SEARCH_DEFAULT_PROXIMITY (-1)
 #endif /* WORD_SEARCH_DEFAULT_PROXIMITY */
 
+class WordTreeArg {
+ public:
+  WordTreeArg(WordList *words, int nuniq, int nproximity, int *document, int document_length, int location) {
+    _words = words;
+    _nuniq = nuniq;
+    _nproximity = nproximity;
+    _document = document;
+    _document_length = document_length;
+    _location = location;
+    _realm = 0;
+  }
+
+  WordList *_words;
+  int _nuniq;
+  int _nproximity;
+  int *_document;
+  int _document_length;
+  int _location;
+  int _realm;
+};
+
 class WordTree : public WordCursorOne {
 public:
   WordTree(WordList* words) :
@@ -78,6 +101,7 @@ public:
   {
     proximity = 0;
     uniq = 0;
+    results = 0;
     verbose = 0;
   }
 
@@ -99,31 +123,37 @@ public:
   // method.
   // Return OK on success, NOTOK on failure.
   //
-  virtual int Prepare(WordList *words, int nuniq, int nproximity, int *document, int document_length, int location) {
+  virtual int Prepare(WordTreeArg& arg) {
     int ret;
-    proximity = nproximity;
-    uniq = nuniq;
-    if((ret = key_semantic.Initialize(document, document_length, location)) != OK)
+    proximity = arg._nproximity;
+    uniq = arg._nuniq;
+    if((ret = key_semantic.Initialize(arg._document, arg._document_length, arg._location, arg._nuniq)) != OK)
       return ret;
     key_semantic.Verbose(verbose);
-    WordKey key(words->GetContext());
+    WordKey key(arg._words->GetContext());
+    //
+    // Set scope
+    //
     if(!scope.empty()) {
       if(key.Set(scope) != OK) {
 	fprintf(stderr, "WordTree::Prepare: setting scope %s failed\n", (char*)scope);
 	return NOTOK;
       }
     }
+    //
+    // Set word
+    //
     if(!search.empty()) {
       unsigned int wordid = WORD_DICT_SERIAL_INVALID;
-      if(words->Dict()->SerialExists(search, wordid) != OK)
+      if(arg._words->Dict()->SerialExists(search, wordid) != OK)
 	return NOTOK;
       key.Set(WORD_KEY_WORD, wordid);
     }
-    return WordCursorOne::Initialize(words, key, 0, 0, HTDIG_WORDLIST_WALKER);
+    return WordCursorOne::Initialize(arg._words, key, 0, 0, HTDIG_WORDLIST_WALKER);
   }
 
-  //-
-  //
+  virtual int WalkNextExclude(const WordKey& key);
+
   virtual int Bounds(const WordKey& bottom, const WordKey& top) = 0;
 
   //-
@@ -135,6 +165,29 @@ public:
     return found;
   }
 
+  //-
+  // Change the search criterion to match realm
+  //
+  virtual int SetRealm(const WordKey& realm) {
+    key_semantic.RealmCopy(realm, GetSearch());
+    return OK;
+  }
+  //-
+  // Undefine the realm
+  //
+  virtual int UndefinedRealm() {
+    key_semantic.RealmUndefined(GetSearch());
+    return OK;
+  }
+  //-
+  // Set the realm to 0
+  //
+  int ClearRealm() {
+    WordKey realm(words->GetContext());
+    key_semantic.RealmClear(realm);
+    return SetRealm(realm);
+  }
+
   //-
   // Returns estimate of the match count.
   //
@@ -154,6 +207,12 @@ public:
   //
   String GetInfo() { return info; }
 
+  //-
+  // Set the list of documents that must be ignored.
+  //
+  virtual inline int SetResults(WordResults* nresults) { results = nresults; return OK; }
+  virtual inline WordResults* GetResults() { return results; }
+
   //-
   // Sort WordTree data members (if any) in ascending frequency order.
   // Return OK on success, NOTOK on failure.
@@ -165,11 +224,19 @@ public:
   // The number of data members deleted is returned in <b>stripped</b>.
   // Return OK on success, NOTOK on failure.
   //
-  virtual int StripNonExistent(unsigned int& stripped) {
+  virtual int StripNonExistent(unsigned int& stripped, unsigned int& killed_mandatory) {
     stripped = 0;
+    killed_mandatory = 0;
     return OK;
   }
 
+  //-
+  // Optimize the search expression. It first asks the expression
+  // (member expr) to optimize itself, then it does some optimization
+  // that cannot be done by the expression itself.
+  //
+  static int TopLevelOptimize(WordTree*& expr);
+
   inline int Verbose(int verbosity) { return verbose = verbosity; }
 
   //
@@ -204,6 +271,10 @@ public:
   // Textual information about the latest match.
   //
   String info;
+  //-
+  // A list of documents that must be ignored
+  //
+  WordResults *results;
   int verbose;
 };
 
@@ -217,7 +288,6 @@ public:
   //
   WordTreeLiteral(WordList* words, const char* string, int string_length, const char* nscope = "") :
     WordTree(words),
-    current_document(words->GetContext()),
     bottom_document(words->GetContext()),
     top_document(words->GetContext())
   {
@@ -254,7 +324,6 @@ public:
   }
 
 protected:
-  WordKey current_document;
   WordKey bottom_document;
   WordKey top_document;
 };
@@ -388,6 +457,17 @@ public:
     return OK;
   }
 
+  virtual int SetResults(WordResults* nresults) {
+    //
+    // Apply to each cursor
+    //
+    unsigned int i;
+    for(i = 0; i < cursors_length; i++)
+      if(cursors[i]->SetResults(nresults) == NOTOK)
+	return NOTOK;
+    return WordTree::SetResults(nresults);
+  }
+
   //-
   // Recursively call WalkInit on each <i>cursors</i>.
   //
@@ -433,14 +513,14 @@ public:
   // Call Prepare on each <i>cursors</i>. Set the <i>search</i> member
   // with an textual representation of the tree starting at this node.
   //
-  virtual int Prepare(WordList *words, int nuniq, int nproximity, int *document, int document_length, int location) {
+  virtual int Prepare(WordTreeArg& arg) {
     int ret;
-    if((ret = WordTree::Prepare(words, nuniq, nproximity, document, document_length, location)) != OK)
+    if((ret = WordTree::Prepare(arg)) != OK)
       return ret;
     unsigned int i;
     for(i = 0; i < cursors_length; i++) {
       cursors[i]->Verbose(verbose);
-      if((ret = cursors[i]->Prepare(words, nuniq, nproximity, document, document_length, location)) != OK)
+      if((ret = cursors[i]->Prepare(arg)) != OK)
 	return ret;
     }
     return Get(search);
@@ -509,12 +589,12 @@ class WordTreeOptional : public WordTreeOperand {
   //
   virtual int Seek(const WordKey& position);
 
-  virtual int Prepare(WordList *words, int nuniq, int nproximity, int *document, int document_length, int location) {
+  virtual int Prepare(WordTreeArg& arg) {
     int ret;
     if((ret = permutation.Initialize(cursors_length, 0, 0, WORD_PERMUTE_PROXIMITY_TOGGLE)) != OK)
       return ret;
     permutation.Verbose(verbose);
-    return WordTreeOperand::Prepare(words, nuniq, nproximity, document, document_length, location);
+    return WordTreeOperand::Prepare(arg);
   }
 
   virtual void SetInfo();
@@ -523,6 +603,26 @@ class WordTreeOptional : public WordTreeOperand {
 
   virtual int UsePermutation() const { return 1; }
 
+  virtual int SetRealm(const WordKey& realm) {
+    unsigned int i;
+    int ret = OK;
+    for(i = 0; i < cursors_length; i++) {
+      if(permutation.Excluded(i)) {
+	//
+	// An excluded cursor is not restricted to the current realm.
+	// The document is excluded no matter where the word appears
+	// in the document.
+	//
+	if((ret = cursors[i]->UndefinedRealm()) != OK)
+	  return ret;
+      } else {
+	if((ret = cursors[i]->SetRealm(realm)) != OK)
+	  return ret;
+      }
+    }
+    return OK;
+  }
+
   virtual int Count(unsigned int& count) const;
 
   //-
@@ -588,21 +688,6 @@ class WordTreeOptional : public WordTreeOperand {
   // 
   //
   int SearchCursorAnd(WordTree& cursor, WordKey& document, WordExclude& permutation);
-  //
-  // We know that : 
-  // 1) document does not contain any excluded words.
-  // 2) contains at least one occurrence of each non excluded word.
-  // The logic, although very similar to WordSearchNear::SearchOne
-  // is therefore simpler. We ignore all excluded cursors and 
-  // return WORD_SEARCH_NOPROXIMITY as soon as a cursor move outside
-  // <document>.
-  //
-  //-
-  // If <b>document</b> contains words that match proximity
-  // requirement, return OK. Return WORD_SEARCH_NOPROXIMITY if proximity
-  // requirement cannot be matched for <document>. 
-  //
-  int CursorsObeyProximity(WordKey& document);
 
   //-
   // Sort the <i>cursors</i> in ascending frequency order using the
@@ -617,7 +702,7 @@ class WordTreeOptional : public WordTreeOperand {
   // deletions in the <b>stripped</i> argument. 
   // Return OK on success, NOTOK on failure.
   //
-  virtual int StripNonExistent(unsigned int& stripped);
+  virtual int StripNonExistent(unsigned int& stripped, unsigned int& killed_mandatory);
 };
 
 // ************************* WordTreeOr implementation ********************
@@ -680,6 +765,14 @@ class WordTreeNear : public WordTreeOptional {
   virtual int UseProximity() const { return WORD_PERMUTE_PROXIMITY_ONLY; }
 
   virtual int AllOrNothing() const { return 1; }
+
+  virtual int Prepare(WordTreeArg& arg) {
+    int proximity_saved = proximity ? proximity : arg._nproximity;
+    int ret = WordTreeOptional::Prepare(arg);
+    proximity = proximity_saved;
+    return ret;
+  }
+
 };
 
 // ************************* WordTreeMandatory implementation ***************
